@@ -1,6 +1,6 @@
 /*
  * (C) Copyright IBM Corporation 2006
- * Copyright (c) 2007, 2009, 2011, 2012, 2013 Oracle and/or its affiliates.
+ * Copyright (c) 2007, 2009, 2011, 2012, 2016 Oracle and/or its affiliates.
  * All Rights Reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -25,6 +25,9 @@
 /*
  * Solaris devfs interfaces
  */
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
 #include <stdlib.h>
 #include <strings.h>
@@ -59,11 +62,10 @@ typedef struct i_devnode {
 } i_devnode_t;
 
 typedef struct nexus {
-    int fd;
     int first_bus;
     int last_bus;
     int domain;
-    char *path;			/* for errors/debugging; fd is all we need */
+    char *path;			/* for open */
     char *dev_path;
     struct nexus *next;
 } nexus_t;
@@ -143,7 +145,6 @@ pci_system_solx_devfs_destroy( void )
 
     for (nexus = nexus_list ; nexus != NULL ; nexus = next) {
 	next = nexus->next;
-	close(nexus->fd);
 	free(nexus->path);
 	free(nexus->dev_path);
 	free(nexus);
@@ -296,14 +297,12 @@ probe_nexus_node(di_node_t di_node, di_minor_t minor, void *arg)
     int pci_node = 0;
     int first_bus = 0, last_bus = PCI_REG_BUS_G(PCI_REG_BUS_M);
     int domain = 0;
-    di_node_t rnode =  DI_NODE_NIL;
 #ifdef __sparc
     int bus_range_found = 0;
     int device_type_found = 0;
     di_prom_prop_t prom_prop;
 #endif
 
-
 #ifdef DEBUG
     nexus_name = di_devfs_minor_path(minor);
     fprintf(stderr, "-- device name: %s\n", nexus_name);
@@ -418,33 +417,24 @@ probe_nexus_node(di_node_t di_node, di_minor_t minor, void *arg)
     if ((fd = open(nexus_path, O_RDWR | O_CLOEXEC)) >= 0) {
 	probe_args_t args;
 
-	nexus->fd = fd;
 	nexus->path = strdup(nexus_path);
 	nexus_dev_path = di_devfs_path(di_node);
 	nexus->dev_path = strdup(nexus_dev_path);
 	di_devfs_path_free(nexus_dev_path);
 
-	if ((rnode = di_init(nexus->dev_path, DINFOCPYALL)) == DI_NODE_NIL) {
-	    (void) fprintf(stderr, "di_init failed: %s\n", strerror(errno));
-	    close(nexus->fd);
-	    free(nexus->path);
-	    free(nexus->dev_path);
-	    free(nexus);
-	    return (DI_WALK_TERMINATE);
-	}
-
 	/* Walk through devices under the rnode */
 	args.pinfo = pinfo;
 	args.nexus = nexus;
 	args.ret = 0;
 
-	(void) di_walk_node(rnode, DI_WALK_CLDFIRST, (void *)&args, probe_device_node);
+	(void) di_walk_node(di_node, DI_WALK_CLDFIRST, (void *)&args, probe_device_node);
+
+	close(fd);
+
 	if (args.ret) {
-	    close(nexus->fd);
 	    free(nexus->path);
 	    free(nexus->dev_path);
 	    free(nexus);
-	    di_fini(rnode);
 	    return (DI_WALK_TERMINATE);
 	}
 
@@ -456,10 +446,6 @@ probe_nexus_node(di_node_t di_node, di_minor_t minor, void *arg)
 	free(nexus);
     }
 
-    if (rnode != DI_NODE_NIL) {
-	di_fini(rnode);
-    }
-
     return DI_WALK_CONTINUE;
 }
 
@@ -552,7 +538,7 @@ pci_device_solx_devfs_probe( struct pci_device * dev )
      * starting to find if it is MEM/MEM64/IO
      * using libdevinfo
      */
-    if ((rnode = di_init(nexus->dev_path, DINFOCPYALL)) == DI_NODE_NIL) {
+    if ((rnode = di_init(nexus->dev_path, DINFOCACHE)) == DI_NODE_NIL) {
 	err = errno;
 	(void) fprintf(stderr, "di_init failed: %s\n", strerror(errno));
     } else {
@@ -724,6 +710,9 @@ pci_device_solx_devfs_map_range(struct pci_device *dev,
 			   strerror(errno));
 	    return err;
 	}
+#ifndef __sparc
+        xsvc_fd = map_fd;
+#endif
     }
 
     map->memory = mmap(NULL, map->size, prot, MAP_SHARED, map_fd, map->base);
@@ -784,6 +773,7 @@ pci_device_solx_devfs_read( struct pci_device * dev, void * data,
     int err = 0;
     unsigned int i = 0;
     nexus_t *nexus;
+    int fd;
 
     nexus = find_nexus_for_bus(dev->domain, dev->bus);
 
@@ -801,11 +791,14 @@ pci_device_solx_devfs_read( struct pci_device * dev, void * data,
     cfg_prg.barnum = 0;
     cfg_prg.user_version = PCITOOL_USER_VERSION;
 
+    if ((fd = open(nexus->path, O_RDWR | O_CLOEXEC)) < 0)
+	return ENOENT;
+
     for (i = 0; i < size; i += PCITOOL_ACC_ATTR_SIZE(PCITOOL_ACC_ATTR_SIZE_1))
     {
 	cfg_prg.offset = offset + i;
 
-	if ((err = ioctl(nexus->fd, PCITOOL_DEVICE_GET_REG, &cfg_prg)) != 0) {
+	if ((err = ioctl(fd, PCITOOL_DEVICE_GET_REG, &cfg_prg)) != 0) {
 	    fprintf(stderr, "read bdf<%s,%x,%x,%x,%llx> config space failure\n",
 		    nexus->path,
 		    cfg_prg.bus_no,
@@ -823,6 +816,8 @@ pci_device_solx_devfs_read( struct pci_device * dev, void * data,
     }
     *bytes_read = i;
 
+    close(fd);
+
     return (err);
 }
 
@@ -838,6 +833,7 @@ pci_device_solx_devfs_write( struct pci_device * dev, const void * data,
     int err = 0;
     int cmd;
     nexus_t *nexus;
+    int fd;
 
     nexus = find_nexus_for_bus(dev->domain, dev->bus);
 
@@ -885,11 +881,17 @@ pci_device_solx_devfs_write( struct pci_device * dev, const void * data,
      */
     cmd = PCITOOL_DEVICE_SET_REG;
 
-    if ((err = ioctl(nexus->fd, cmd, &cfg_prg)) != 0) {
+    if ((fd = open(nexus->path, O_RDWR | O_CLOEXEC)) < 0)
+	return ENOENT;
+
+    if ((err = ioctl(fd, cmd, &cfg_prg)) != 0) {
+	close(fd);
 	return (err);
     }
     *bytes_written = size;
 
+    close(fd);
+
     return (err);
 }
 
@@ -911,6 +913,7 @@ pci_device_solx_devfs_open_legacy_io(struct pci_io_handle *ret,
     if (sysi86(SI86V86, V86SC_IOPL, PS_IOPL) == 0) {
 	ret->base = base;
 	ret->size = size;
+	ret->is_legacy = 1;
 	return ret;
     }
 #endif
@@ -1067,7 +1070,7 @@ pci_system_solx_devfs_create( void )
 	return 0;
     }
 
-    if ((di_node = di_init("/", DINFOCPYALL)) == DI_NODE_NIL) {
+    if ((di_node = di_init("/", DINFOCACHE)) == DI_NODE_NIL) {
 	err = errno;
 	(void) fprintf(stderr, "di_init() failed: %s\n",
 		       strerror(errno));
